#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <float.h>
#include <math.h>
#include <errno.h>


#include "Light.h"
#include "LightPassband.h"

struct model_flux {
  float tmin; /* inclusive */
  float tmax; /* exclusive */
  float log_g;
  unsigned int g_index;

  unsigned int    m_grid_n;
  unsigned int    m_grid_nmax;
  float         * m_grid_t;
  float         * m_grid_f;

  struct model_flux * next;
};
  
struct passband {
  char name[NF_PB_NMAX];
  int  external;
  float wavelength;
  float m2nd;
  unsigned int  ld_types;

  unsigned int  ld_grid_n[NF_PB_GMAX][NF_LDMAX];
  unsigned int  ld_grid_nmax[NF_PB_GMAX][NF_LDMAX];/* allocated               */
  float *       ld_grid_t[NF_PB_GMAX][NF_LDMAX];
  float *       ld_grid_ldc[NF_PB_GMAX][NF_LDMAX]; /* log g = 3.5, 4, 4.5, 5  */
	 
  float         t_upper[NF_PB_GMAX];
  float         t_lower[NF_PB_GMAX];
  double        bb_conv_upper[NF_PB_GMAX]; /* BB * Conv = ModelFlux */
  double        bb_conv_lower[NF_PB_GMAX];
  
  struct model_flux * mlist;
};

static struct passband pb_bands[NUM_MAG] = { 0 };

static float LogG[NF_PB_GMAX] = { 3.5, 4.0, 4.5, 5.0 };

static char *LDModel[NF_LDMAX] = { "lin", "quad", "sqrt", "four" };
static int   scan_expect[NF_LDMAX] = { 3, 4, 4, 6 };

#include <stdarg.h>
static void perr(const char * fmt, ...)
{
  va_list ap;
  fputs ("ERROR: ", stderr);
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  putc ('\n', stderr);
  return;
}

char  * nf_pb_get_name(unsigned int pos)
{
  assert(pos < NUM_MAG);
  return pb_bands[pos].name;
}

float   nf_pb_get_wave(unsigned int pos)
{
  assert(pos < NUM_MAG);
  return pb_bands[pos].wavelength;
}

void    nf_pb_set_wave(unsigned int pos, float wave)
{
  assert(pos < NUM_MAG);
  pb_bands[pos].wavelength = wave;
  return;
}

float   nf_pb_get_m2nd(unsigned int pos)
{
  assert(pos < NUM_MAG);
  return pb_bands[pos].m2nd;
}

unsigned int nf_pb_has_ldType(int pos, ldType type)
{
  assert(pos < NUM_MAG);
  return (pb_bands[pos].ld_types & (unsigned int)type);
}

void    nf_pb_updateNames(char filter[][NF_PB_NMAX])
{
  int i;
  for (i = 0; i < NUM_MAG; ++i) {
    strncpy(filter[i], nf_pb_get_name(i), NF_PB_NMAX);
    filter[i][NF_PB_NMAX-1] = '\0'; }
  return;
}

/*                                       */
/* ---- Limb Darkening Coefficients ---- */
/*                                       */

static int indexOfLogG (float log_g)
{
  int i;
  static float last_g = -1.0;
  static int   last_i = -1;
  
  assert (log_g > 0.0);

  if (last_g == log_g)
    {
      return last_i;
    }
  else
    {
      last_i = -1;
      
      for (i = 0; i < NF_PB_GMAX; ++i)
	{
	  float lg = LogG[i];
	  if      (fabs(log_g - lg) < (log_g+lg) * FLT_EPSILON)
	    {
	      last_g = log_g;
	      last_i = i;
	      return i;
	    }
	}
    }

  perr("invalid log g %f", log_g);
  assert( (last_i >= 0) && (last_i < NF_PB_GMAX) );
  return 0; /* fail safe */
}

static int findWithin(float key, float * r, int n)
{
  static int i = 0;
  static int high, low;
  
  /* speedup (?)
   */
  if (i < n && key >= r[i] && key < r[i+1])
    {
      return ( i );
    }

  /* return 0 if < r[0]; 
   */
  for ( low = (-1), high = n;  high-low > 1;  )
    {
      i = (high+low) / 2;
      if ( key < r[i])  
	high = i;
      else             
	low  = i;
    }

  /* make binary search stable by explicitely selecting the low value
   */
  i = (low < 0) ? 0 : low;

  return i;
}

/* Caller must ensure that 'array' is big enough for ldType;
 * Returns: pointer to 'array', or NULL on failure 
 */
float * nf_pb_fill_ldArray(int pos, ldType ld_type, float log_g, float t,
			   float * array)
{
  assert(pos < NUM_MAG);

  if (0 == (pb_bands[pos].ld_types & (unsigned int)ld_type))
    {
      return NULL;
    }

  else
    {
      int i = indexOfLogG(log_g);

      if (i < 0)
	return NULL;
      
      if (ld_type == LD_LINEAR)
	{
	  int j = findWithin(t, pb_bands[pos].ld_grid_t[i][LD_LN], pb_bands[pos].ld_grid_n[i][LD_LN]);
	  array[0] = pb_bands[pos].ld_grid_ldc[i][LD_LN][j];
	}
      else if (ld_type == LD_SQR)
	{
	  int j = findWithin(t, pb_bands[pos].ld_grid_t[i][LD_QD], pb_bands[pos].ld_grid_n[i][LD_QD]);
	  array[0] = pb_bands[pos].ld_grid_ldc[i][LD_QD][2*j];
	  array[1] = pb_bands[pos].ld_grid_ldc[i][LD_QD][2*j+1];
	}
      else if (ld_type == LD_SQRT)
	{
	  int j = findWithin(t, pb_bands[pos].ld_grid_t[i][LD_SQ], pb_bands[pos].ld_grid_n[i][LD_SQ]);
	  array[0] = pb_bands[pos].ld_grid_ldc[i][LD_SQ][2*j];
	  array[1] = pb_bands[pos].ld_grid_ldc[i][LD_SQ][2*j+1];
	}
      else
	return NULL;
    }
  return array;
}


/* ------------------------------------*/
/*                                     */
/*  Model Fluxes                       */
/*                                     */
/* ------------------------------------*/

/* Return bb adjustment value (flux = bb_flux * adjust)
 */
double   nf_bb_adjust(int pos, float log_g, float t)
{
  static float last_g = 0.0;
  static int   g_index;
  
  if (log_g == last_g) /* do nothing */;
  else
    {
      g_index = indexOfLogG (log_g);
      last_g  = log_g;
    }

  if (t <= pb_bands[pos].t_lower[g_index])
    return pb_bands[pos].bb_conv_lower[g_index];
  else
    return pb_bands[pos].bb_conv_upper[g_index];
}   
    
  
/* Returns size of array(s), or 0 on failure
 */
unsigned int  nf_modelFlux(int pos, float log_g, float t,
		     float ** tarr, float ** farr)
{

  static struct model_flux * last_md[NUM_MAG] = { NULL };
  static float last_g[NUM_MAG] = { 0.0 };
  static float last_tlower[NUM_MAG] = { 0.0 };
  static float last_tupper[NUM_MAG] = { 0.0 };
  
  struct passband pb = pb_bands[pos];
  
  if (last_md[pos])
    {
      if ( (log_g == last_g[pos]) &&
	   (last_tlower[pos] <= t) &&  (t < last_tupper[pos]) )
	{
	  *tarr = last_md[pos]->m_grid_t;
	  *farr = last_md[pos]->m_grid_f;
	  return last_md[pos]->m_grid_n;
	}
    }
  
  last_md[pos] = pb.mlist;
  
  while (last_md[pos])
    {
      if (fabs(log_g - last_md[pos]->log_g) <=
	  (fabs(log_g) + fabs(last_md[pos]->log_g)) * FLT_EPSILON)
	{
	  if (t >= last_md[pos]->tmin && t < last_md[pos]->tmax)
	    {
	      *tarr = last_md[pos]->m_grid_t;
	      *farr = last_md[pos]->m_grid_f;
	      last_g[pos]      = log_g;
	      last_tlower[pos] = last_md[pos]->tmin;
	      last_tupper[pos] = last_md[pos]->tmax;
	      return last_md[pos]->m_grid_n;
	    }
	}
      last_md[pos] = last_md[pos]->next;
    }

  *tarr = NULL;
  *farr = NULL;
  return 0;
}

float nf_interpolate(float t, float * tarr, float * farr, int len, int * where)
{
  double f0, f1, f2;
  double t0, t1, t2;
  double result;

  if (len > 0)
    {
      int hkey = findWithin(t, tarr, len);
      
      hkey  = (hkey < (len-2)) ? hkey : (len - 3);
      
      *where = hkey;
      
      f0    = farr[hkey];
      f1    = farr[hkey+1];
      f2    = farr[hkey+2];
      
      t0    = tarr[hkey];
      t1    = tarr[hkey+1];
      t2    = tarr[hkey+2];
      
      /* ------------- do a parabolic interpolation -----------------    
       * (uses Lagrange interpolation polynome)
       */
      
      result = ( f0  * ((t-t1)*(t-t2)) / ((t0-t1)*(t0-t2))
		 +    f1  * ((t-t0)*(t-t2)) / ((t1-t0)*(t1-t2))
		 +    f2  * ((t-t0)*(t-t1)) / ((t2-t0)*(t2-t1)) );
      
      if      (f0 <= result && result <= f1)
	return result;
      else if (f0 >= result && result >= f1) 
	return result;
      else {
	/* linear interpolation */
	float x = (t - t0) / (t1 - t0);
	result = ((1.0 - x) * f0 + x * f1);
	return result;
      }
    }
  else
    {
      *where = -1;
      /* len == 0, i.e. flux array not found */
      return (float)0.0;
    }
}


/* ------------------------------------*/
/*                                     */
/* Data Read                           */
/*                                     */
/* ------------------------------------*/

enum vItems 
  {
    VAL_NAME = (1 << 0),
    VAL_LDC  = (1 << 1),
    VAL_FLUX = (1 << 2),
    VAL_WAVE = (1 << 3),
    VAL_M2ND = (1 << 4)
  };

static char * strip(char * line)
{
  while (*line == ' ' || *line == '\t' || *line == '\n') ++line;
  return line;
}

int fetchName(int pos, unsigned int * items, char * str, FILE * fd,
	      char * line, size_t len)
{
  char * buf;
  
  if (1 != sscanf(str, "NAME %ms", &buf))
    {
      perr("invalid line %s", str);
      return -1;
    }

  if (strlen(buf) > (NF_PB_NMAX-1))
    {
      perr("name (%s) too long, maximum is %d", buf, (NF_PB_NMAX-1)); 
      return -1;
    }

  strcpy(pb_bands[pos].name, buf);
  *items |= VAL_NAME;

  if (NULL == fgets( line, len, fd ))
    line[0] = '\0';

  free(buf);
  return 0;
}

float fetchFloat(char * format, char * str, FILE * fd, char * line, size_t len)
{
  float value;
  
  if (1 != sscanf(str, format, &value))
    {
      perr("invalid line %s", str);
      return -1.0;
    }

  if (value < 0)
    {
      perr("invalid value (%f)", value);
      return -1.0;
    }

  if (NULL == fgets( line, len, fd ))
    line[0] = '\0';
  return value;
}

int fetchWave(int pos, unsigned int * items, char * str, FILE * fd,
	      char * line, size_t len)
{
  float wavel = fetchFloat ("WAVELENGTH %f", str, fd, line, len);

  if (wavel < 0.0)
    return -1;

  pb_bands[pos].wavelength = wavel;
  *items |= VAL_WAVE;
  return 0;
}

int fetchM2nd(int pos, unsigned int * items, char * str, FILE * fd,
	      char * line, size_t len)
{
  float m2nd = fetchFloat ("CORR_FACTOR %f", str, fd, line, len);
  if (m2nd < 0.0)
    return -1;

  pb_bands[pos].m2nd = m2nd;
  *items |= VAL_M2ND;
  return 0;
}

static void extendLDC(int pos, int g, int ldc)
{
  unsigned int size;
  unsigned int mul = scan_expect[ldc] - 2;
  
  size = pb_bands[pos].ld_grid_nmax[g][ldc] + 100;

  pb_bands[pos].ld_grid_t[g][ldc]    = realloc(pb_bands[pos].ld_grid_t[g][ldc],
					       size * sizeof(float));
  pb_bands[pos].ld_grid_ldc[g][ldc]  =
    realloc(pb_bands[pos].ld_grid_ldc[g][ldc], mul * size * sizeof(float));
  
  pb_bands[pos].ld_grid_nmax[g][ldc] = size;

  return;
}

static int ldcIndex(char * str)
{
  char  ldc[32];
  int   i;
  
  if (1 != sscanf(str, "LDC_START %31s", ldc))
    {
      perr("invalid line %s", str);
      return -1;
    }

  for (i = 0; i < NF_LDMAX; ++i)
    if (0 == strncmp(ldc, LDModel[i], strlen( LDModel[i] )))
      return i;

  perr("invalid line %s", str);
  return -1;
}

int fetchLDC (int pos, unsigned int * items, char * str, FILE * fd,
	      char * line, size_t len)
{
  int j, i;
  int ldc = ldcIndex(str);

  if (ldc < 0)
    return -1;

  while( NULL != fgets( line, len, fd ))
    {
      float t, g, f[NF_PB_LMAX];
      int   ret;
      int   npara;
      int   k;
      
      str = strip(line);         /* remove leading whitespace */
      if (*str == '#' || *str == '\0') continue; /* comment|empty */

      ret = sscanf(str, "%f %f %f %f %f %f", &t, &g, &f[0], &f[1], &f[2], &f[3]);

      if (ret != scan_expect[ldc])
	break;

      npara = scan_expect[ldc] - 2;
      
      i = indexOfLogG(g);
      if (i < 0)
	return -1;
      
      if ( (pb_bands[pos].ld_grid_n[i][ldc] > 0) &&
	   (t < pb_bands[pos].ld_grid_t[i][ldc][0]) )
	{
	  perr("unsorted LDC data %s", str);
	  return -1;
	}
      
      if (pb_bands[pos].ld_grid_n[i][ldc] == pb_bands[pos].ld_grid_nmax[i][ldc])
	extendLDC(pos, i, ldc);
      
      assert(pb_bands[pos].ld_grid_nmax[i][ldc] > pb_bands[pos].ld_grid_n[i][ldc]);
	     
      j = pb_bands[pos].ld_grid_n[i][ldc];
      pb_bands[pos].ld_grid_t[i][ldc][j]   = t;

      for (k = 0; k < npara; ++k)
	{
	  pb_bands[pos].ld_grid_ldc[i][ldc][npara*j + k]    = f[k];
	}
      
      ++pb_bands[pos].ld_grid_n[i][ldc];
    }

  j = 0;
  for (i = 0; i < NF_PB_GMAX; ++i)
    if (pb_bands[pos].ld_grid_n[i][ldc] > 0) ++j;
  if (j == 0)
    {
      perr("no valid LDC data");
      return -1;
    }
    
  pb_bands[pos].ld_types |= (1 << ldc);

  *items |= VAL_LDC;
  return 0;
}

static struct model_flux * newModel(char * str)
{
  float log_g, f_tmin, f_tmax;
  int g_index;
  struct model_flux * mFlux = NULL;
  
  if (3 != sscanf(str, "FLUX_START %f %f %f", &log_g, &f_tmin, &f_tmax))
    {
      perr("invalid line %s", str);
      return NULL;
    }
  
  g_index = indexOfLogG(log_g);
  if ( (g_index < 0) || (f_tmin < 0) || (f_tmax < 0) || (f_tmin >= f_tmax) )
    {
      perr("invalid line %s", str);
      return NULL;
    }
  mFlux = calloc(1, sizeof(struct model_flux));

  mFlux->log_g    = log_g;  mFlux->g_index     = g_index;
  mFlux->tmin     = f_tmin; mFlux->tmax        = f_tmax;
  mFlux->m_grid_n = 0;      mFlux->m_grid_nmax = 0;
  mFlux->m_grid_t = NULL;   mFlux->m_grid_f    = NULL;

  return mFlux;
}

static void extendFlux(struct model_flux * mFlux)
{
  unsigned int size = mFlux->m_grid_nmax + 100;

  mFlux->m_grid_t = realloc(mFlux->m_grid_t, size*sizeof(float));
  mFlux->m_grid_f = realloc(mFlux->m_grid_f, size*sizeof(float));

  mFlux->m_grid_nmax = size;
  return;
}

int fetchFlux (int pos, unsigned int * items, char * str, FILE * fd,
	       char * line, size_t len)
{
  struct model_flux * mFlux = newModel(str);

  if (!mFlux)
    return -1;
  
  while( NULL != fgets( line, len, fd ))
    {
      float t, flux;
      
      str = strip(line);         /* remove leading whitespace */
      if (*str == '#' || *str == '\0') continue; /* comment|empty */

      if (2 != sscanf(str, "%f %f", &t, &flux))
	break;

      if ((mFlux->m_grid_n > 0) && (t < mFlux->m_grid_t[0]))
	{
	  perr("unsorted flux data (log g = %f, tmin = %f)",
	       mFlux->log_g, mFlux->tmin);
	  return -1;
	}
      
      if (mFlux->m_grid_n == mFlux->m_grid_nmax)
	extendFlux(mFlux);

      mFlux->m_grid_t[mFlux->m_grid_n] = t;
      mFlux->m_grid_f[mFlux->m_grid_n] = flux;
      
      ++mFlux->m_grid_n;
    }

  if (mFlux->m_grid_n == 0)
    {
      perr("no valid flux data (log g = %f, tmin = %f)",
	   mFlux->log_g, mFlux->tmin);
      free(mFlux);
      return -1;
    }
  
  mFlux->next         = pb_bands[pos].mlist;
  pb_bands[pos].mlist = mFlux;
  *items |= VAL_FLUX;
  return 0;
}

static void finishModelLimits(int pos, int i, struct model_flux * mFlux)
{
  if ((mFlux->tmin < pb_bands[pos].t_lower[i]) ||
      (pb_bands[pos].t_lower[i] == 0))
    pb_bands[pos].t_lower[i] = mFlux->tmin;
  
  if (mFlux->tmax > pb_bands[pos].t_upper[i])
    pb_bands[pos].t_upper[i] = mFlux->tmax;
  
  return;
}

double nf_bb_flux(float t, float wavel, float m2nd)
{
  static  double  Wien =  2897.756;   /* constant for Wien's law          */
  static  double  CC1 = 3.7418498e+15;/* 10e(5*4)      erg cm / sec       */
  static  double  CC2 = 14388.3361;   /* 10e4          cm K               */
  /* B = CC1/WaveL^5 * 1/(exp(CC2/(T*WaveL))-1) WaveL(ength) in _cm_  */

  double  lmax = Wien/t;              /* Blackbody peak                   */
  double  ewav = wavel + 5 * (lmax-wavel) * m2nd/SQR(wavel);

  /* return (CC1/(wavel*wavel*wavel*wavel*wavel)) * 1.0/(exp(CC2/(wavel * t))-1.0); */
  return (CC1/(ewav*ewav*ewav*ewav*ewav)) * 1.0/(exp(CC2/(ewav * t))-1.0);
}

static int finishModelBB(int pos, int g_index)
{
  float tmin = pb_bands[pos].t_lower[g_index];
  float tmax = pb_bands[pos].t_upper[g_index] - 0.1;
  
  float * tarr = NULL;
  float * farr = NULL;
  int     n;
  double  flux;
  double  bbf;
  int     ignore;

  n = nf_modelFlux(pos, LogG[g_index], tmin, &tarr, &farr);
  if (n && tarr && farr)
    {
      flux = nf_interpolate(tmin, tarr, farr, n, &ignore);
      bbf = nf_bb_flux(tmin, pb_bands[pos].wavelength, pb_bands[pos].m2nd);
      pb_bands[pos].bb_conv_lower[g_index] = flux / bbf;
    }
  else
    return -1;
  
  n = nf_modelFlux(pos, LogG[g_index], tmax, &tarr, &farr);
  if (n && tarr && farr)
    {
      flux = nf_interpolate(tmax, tarr, farr, n, &ignore);
      bbf  = nf_bb_flux(tmax, pb_bands[pos].wavelength, pb_bands[pos].m2nd);
      pb_bands[pos].bb_conv_upper[g_index] = flux / bbf;
    }
  else
    return -1;
  return 0;
}

static int finishModel(int pos)
{
  int i;
  struct model_flux * mFlux;
  
  for (i = 0; i < NF_PB_GMAX; ++i)
    {
      mFlux = pb_bands[pos].mlist;
      
      while (mFlux)
	{
	  if (mFlux->g_index == i)
	    {
	      finishModelLimits(pos, i, mFlux);
	    }
	  mFlux = mFlux->next;
	}
      
      if (pb_bands[pos].t_upper[i] == pb_bands[pos].t_lower[i])
	{
	  /* No flux for this log_g */
	  pb_bands[pos].bb_conv_upper[i] = 1.0;
	  pb_bands[pos].bb_conv_lower[i] = 1.0;
	}
      else
	{
	  if (0 != finishModelBB(pos, i))
	    {
	      perr("failed to compute BB conversion factors");
	      return -1;
	    }
	}
    }
  
  return 0;
}

static int readFile(int pos, FILE * fd)
{
  char   line[1024];
  unsigned int items = 0;

  while( NULL != fgets( line, sizeof(line), fd ))
    {
      char * str = strip(line);      /* remove leading whitespace */
      if (*str == '#' || *str == '\0') continue; /* comment|empty */

      do
	{
	  if      (0 == strncmp(str, "NAME", strlen("NAME")))
	    {
	      if (0 != fetchName(pos, &items, str, fd, line, sizeof(line)))
		{
		  perr("failed to read passband name");
		  return -1;
		}
	    }
	  else if (0 == strncmp(str, "WAVELENGTH", strlen("WAVELENGTH")))
	    {
	      if (0 != fetchWave (pos, &items, str, fd, line, sizeof(line)))
		{
		  perr("failed to read passband wavelength");
		  return -1;
		}
	    }
	  else if (0 == strncmp(str, "CORR_FACTOR", strlen("CORR_FACTOR")))
	    {
	      if (0 != fetchM2nd (pos, &items, str, fd, line, sizeof(line)))
		{
		  perr("failed to read passband correction factor");
		  return -1;
		}
	    }
	  else if (0 == strncmp(str, "LDC_START", strlen("LDC_START")))
	    {
	      if (0 != fetchLDC (pos, &items, str, fd, line, sizeof(line)))
		{
		  perr("failed to read LDC data");
		  return -1;
		}
	    }
	  else if (0 == strncmp(str, "FLUX_START", strlen("FLUX_START")))
	    {
	      if (0 != fetchFlux(pos, &items, str, fd, line, sizeof(line)))
		{
		  perr("failed to read flux data");
		  return -1;
		}
	    }
	  else
	    {
	      break;
	    }
	  str = strip(line); if (*str == '#') break;
	}
      while (*str);
    }

  if ( (items | VAL_FLUX) != (VAL_NAME | VAL_LDC | VAL_WAVE | VAL_M2ND | VAL_FLUX ))
    {
      perr("some item(s) [name/wavelength/correction factor/LDC] is/are missing");
      return -1;
    }
  else if (0 == (items & VAL_FLUX))
    {
      WARNING(_("No flux data found in file"));
    }
  
  if (0 != finishModel(pos))
    return -1;

  strncpy(Filters[pos], pb_bands[pos].name, NF_PB_NMAX);

  pb_bands[pos].external = 1;
  return 0;
}

#if 0

static double BBcalc (double lambda, double t)
{
  static  double  CC1 = 3.7418498e+15;/* 10e(5*4)      erg cm / sec       */
  static  double  CC2 = 14388.3361;   /* 10e4          cm K               */
                                      /* WaveL in 10e4 cm                 */
  /* we compute I(1) -- Intensity normal to surface                   */
  /*  in blackbody approximation                                      */  
  /* B = CC1/WaveL^5 * 1/(exp(CC2/(T*WaveL))-1) WaveL(ength) in _cm_  */

  return ( (CC1/(lambda * lambda * lambda * lambda * lambda))
	   * 1.0 / (exp( CC2/(lambda * t))-1.0) );
}

static void checkFile(int pos)
{
  int i, j, k;
  float * res;
  float   array[2];
  float *tarr, *farr;
  int    len;
  float  flux = 1.0;
  
  if (0 != strcmp(pb_bands[pos].name, "UVM2"))
    fprintf(stderr, "FAIL: name %s != UVM2\n", pb_bands[pos].name);
  if (pb_bands[pos].wavelength != (float)0.231)
    fprintf(stderr, "FAIL: wavelength %f != 0.231\n", pb_bands[pos].wavelength);
  if (pb_bands[pos].m2nd != (float)0.00646)
    fprintf(stderr, "FAIL: corr_factor %f != 0.00646\n", pb_bands[pos].m2nd);
  
  for (i = 0; i < NF_PB_GMAX; ++i)
    for (k = 0; k < NF_LDMAX; ++k)
      if (pb_bands[pos].ld_grid_n[i][k] != (6+i))
	fprintf(stderr, "FAIL: n_ld[%f][%d] %d != %d\n", LogG[i], k, pb_bands[pos].ld_grid_n[i][k], (6+i));

  for (i = 0; i < NF_PB_GMAX; ++i)
    for (k = 0; k < NF_LDMAX; ++k)
      if (pb_bands[pos].ld_grid_t[i][k][0] != (float)2000.0)
	fprintf(stderr, "FAIL: t_ld[%f][%d] %f != %f\n", LogG[i], k, pb_bands[pos].ld_grid_t[i][k][0], 2000.0);

  for (i = 0; i < NF_PB_GMAX; ++i) {
    for (k = 0; k < NF_LDMAX; ++k) {
      j = pb_bands[pos].ld_grid_n[i][k]-1;
      if (fabs(pb_bands[pos].ld_grid_t[i][k][j] - (7000.0+i*1000.0)) > 10*FLT_EPSILON*10000.0 )
	fprintf(stderr, "FAIL: t_ld[%f][%d] %f != %f\n", LogG[i], k, pb_bands[pos].ld_grid_t[i][k][j], 7000.0+i*1000.0);
    }
  }

  res = nf_pb_fill_ldArray(pos, LD_LINEAR, 3.5, 3000, array);
  if (!res || ( res[0] != (float) 0.5 ))
    fprintf(stderr, "FAIL g = %f, t = %f, res = %f\n", 3.5, 3000.0, (res == NULL) ? -1.0 : res[0]);
  
  res = nf_pb_fill_ldArray(pos, LD_LINEAR, 4.0, 2000, array);
  if (!res || ( res[0] != (float) 0.2 ))
    fprintf(stderr, "FAIL g = %f, t = %f, res = %f\n", 4.0, 2000.0, (res == NULL) ? -1.0 : res[0]);
  
  res = nf_pb_fill_ldArray(pos, LD_LINEAR, 4.0, 3000, array);
  if (!res || ( res[0] != (float) 0.3 ))
    fprintf(stderr, "FAIL g = %f, t = %f, res = %f\n", 4.0, 3000.0, (res == NULL) ? -1.0 : res[0]);
  
  res = nf_pb_fill_ldArray(pos, LD_SQR, 4.5, 4100, array);
  if (!res || ( res[0] != (float) 0.1 ) || ( res[1] != (float) 0.4 ))
    fprintf(stderr, "FAIL g = %f, t = %f, res = %f, %f\n", 4.5, 4100.0,
	    (res == NULL) ? -1.0 : res[0], (res == NULL) ? -1.0 : res[1]);
  
  res = nf_pb_fill_ldArray(pos, LD_SQRT, 4.5, 4100, array);
  if (!res || ( res[0] != (float) 0.8 ) || ( res[1] != (float) 0.9 ))
    fprintf(stderr, "FAIL g = %f, t = %f, res = %f, %f\n", 4.5, 4100.0,
	    (res == NULL) ? -1.0 : res[0], (res == NULL) ? -1.0 : res[1]);
  
  res = nf_pb_fill_ldArray(pos, LD_SQRT, 5.0, 7200, array);
  if (!res || ( res[0] != (float) 0.81 ) || ( res[1] != (float) 0.45 ))
    fprintf(stderr, "FAIL g = %f, t = %f, res = %f, %f\n", 5.0, 7200.0,
	    (res == NULL) ? -1.0 : res[0], (res == NULL) ? -1.0 : res[1]);
  
  res = nf_pb_fill_ldArray(pos, LD_SQRT, 5.0, 1200, array);
  if (!res || ( res[0] != (float) 0.32 ) || ( res[1] != (float) 0.42 ))
    fprintf(stderr, "FAIL g = %f, t = %f, res = %f, %f\n", 5.0, 1200.0,
	    (res == NULL) ? -1.0 : res[0], (res == NULL) ? -1.0 : res[1]);
  
  res = nf_pb_fill_ldArray(pos, LD_SQRT, 5.0, 9900, array);
  if (!res || ( res[0] != (float) 0.39 ) || ( res[1] != (float) 0.49 ))
    fprintf(stderr, "FAIL g = %f, t = %f, res = %f, %f\n", 5.0, 1200.0,
	    (res == NULL) ? -1.0 : res[0], (res == NULL) ? -1.0 : res[1]);
  
  res = nf_pb_fill_ldArray(pos, LD_SQRT, 5.0, 10001, array);
  if (!res || ( res[0] != (float) 0.31 ) || ( res[1] != (float) 0.41 ))
    fprintf(stderr, "FAIL g = %f, t = %f, res = %f, %f\n", 5.0, 1200.0,
	    (res == NULL) ? -1.0 : res[0], (res == NULL) ? -1.0 : res[1]);

  len = nf_modelFlux(pos, (float)3.5, (float) 2300.0,
		     &tarr, &farr);
  if (len == 0) 
    fprintf(stderr, "FAIL flux array g = %f, t = %f\n", 3.5, 2300.0);
  else
    flux = nf_interpolate(2300.0, tarr, farr, len, &k);
  if (fabs( flux - 6.70707304e+08) > 10*flux*FLT_EPSILON)
    fprintf(stderr, "FAIL flux g = %f, t = %f, %e <> %e\n", 3.5, 2300.0, flux, 6.70707304e+08);
    
  len = nf_modelFlux(pos, (float)4.5, (float) 5000.0,
		     &tarr, &farr);
  if (len == 0) 
    fprintf(stderr, "FAIL flux array g = %f, t = %f\n", 4.5, 5000.0);
  else
    flux = nf_interpolate(5000.0, tarr, farr, len, &k);
  if (fabs( flux - 4.270032e+13) > 10*flux*FLT_EPSILON)
    fprintf(stderr, "FAIL flux g = %f, t = %f, %e <> %e\n", 4.5, 5000.0, flux, 4.270032e+13);

  flux = BBcalc(pb_bands[pos].wavelength, 12100.0) *  nf_bb_adjust(pos, 4.5, 12100.0);
  if (fabs( flux - 4.403863e+17) > 10*flux*FLT_EPSILON)
    fprintf(stderr, "FAIL flux g = %f, t = %f, %e <> %e\n", 4.5, 12100.0, flux, 4.403863e+17);
      
  fprintf(stderr, "Check finished!\n");
  return;
}
#endif

/* Exits on failure. fileName is "pos:path"
 */
void     nf_pb_readFile(const char * fileName)
{
  FILE * fd;
  char   cpos;
  int    pos = -1;
  char * path;
  int    n, i;
  int    result;
  char   code[NUM_MAG+1] = "UBVRIJHKuvby";
  
  assert(fileName != NULL);

  n = sscanf(fileName, "%c:%ms", &cpos, &path);
  if (n == 2)
    for (i = 0; i < NUM_MAG; ++i)
      if (cpos == code[i]) pos = i;
  
  if ( (n != 2) || (pos < 0) || (pos >= NUM_MAG) )
    {
      char * tmp = malloc(64 + strlen(fileName));
      if (n != 2)
	snprintf(tmp, 64 + strlen(fileName), "%s %s",
		 _("Replace Filter: Invalid argument:"), fileName);
      else
	snprintf(tmp, 64 + strlen(fileName), "%s %c",
		 _("Replace Filter: Invalid filter code:"), cpos);
      nf_error(tmp); /* exits */
    }

  fd = fopen(path, "r");
  if (!fd)
    {
      char * e = strerror(errno);
      char * tmp = malloc(64 + strlen(e) + strlen(path) + 3);
      snprintf(tmp, 64 + strlen(e) + strlen(path) + 3, "Replace Filter: %s: %s", e, path);
      nf_error(tmp); /* exits */
    }

  result = readFile(pos, fd);

#if 0
  checkFile(pos);
#endif
  
  if (0 != result)
    {
      char * tmp = malloc(64 + strlen(path));
      snprintf(tmp, 64 + strlen(path), "%s %s",
	       _("Replace Filter: Error while reading file:"), path);
      nf_error(tmp); /* exits */
    }

  fclose(fd);
  free (path);
  return;
}

/*****************************************************************************/
/*                                                                           */
/* Read Internal Data                                                        */
/*                                                                           */
/*****************************************************************************/

/* Wavelength in micrometer (1000 A = 0.1 mu) */
static float pb_wavel[NUM_MAG] = {   
  0.3585, 0.4383, 0.5502,
  0.6400, 0.7900,
  1.25,   1.62,   2.2,
  0.3500, 0.4110, 0.4670, 0.5470
};

/* (second moments of passbands)^2 / FWHM                              */
/* approximated following the prescription by                          */
/* Young A.T. 1992, Astronomy and Astrophysics 257, 366; Equation 12   */
/* m2^2 = [FW(0.05)/4.4]^2                                             */

static float  pb_m2nd[NUM_MAG]   = {
   0.000448,  0.001236,  0.001405,
   4.556e-3,  1.845e-3,  
     0.0103,    0.0211,     0.023,
  1.8061e-5, 1.3514e-5, 1.6357e-5, 2.8619e-5
};


static char * pb_name[NUM_MAG] = {
  "U Band", "B Band", "V Band", "R Band", "I Band",
  "J Band", "H Band", "K Band",
  "Stroemgren u", "Stroemgren v", "Stroemgren b", "Stroemgren y"
};

/*  stellar temperature grid for limb darkening coefficients              */

static   float  TH[] = {   
  2000., 2200., 2400., 2600., 2800., 3000., 3200., 3400., 3600., 3800., 
  4000., 4200., 4400., 4600., 4800., 5000., 5200., 5400., 5600., 5800., 
  7000., 7200., 7400., 7600., 7800., 8000., 8200., 8400., 8600., 8800., 
  9000., 9200., 9400., 9600., 9800.};

static   float  TK[] = {   
  3500.,  3750.,  4000.,  4250.,  4500.,  4750.,  5000.,  5250., 5500., 
  5750.,  6000.,  6250.,  6500.,  6750.,  7000.,  7250.,  7500., 7750., 
  8000.,  8250.,  8500.,  8750.,  9000.,  9250.,  9500.,  9750., 10000., 
  10250., 10500., 10750., 11000., 11250., 11500., 11750., 12000., 12250., 
  12500., 12750., 13000., 14000., 15000., 16000., 17000., 18000., 19000., 
  20000., 21000., 22000., 23000., 24000., 25000., 26000., 27000., 28000., 
  29000., 30000., 31000., 32000., 33000., 34000., 35000., 36000., 37000., 
  38000., 39000., 40000., 41000., 42000., 43000., 44000., 45000., 46000.,
  47000., 48000., 49000., 50000 };

#include "LDC_PHOENIX_3.5.h"
#include "LDC_PHOENIX_4.0.h"
#include "LDC_PHOENIX_4.5.h"
#include "LDC_PHOENIX_5.0.h"

/* For surface gravities not extending to 50000 K, the last value
   is repeated to fill the tables. */ 
#include "LDC_ATLAS_3.5.h"
#include "LDC_ATLAS_4.0.h"
#include "LDC_ATLAS_4.5.h"
#include "LDC_ATLAS_5.0.h"

/* Kurucz ATLAS Intensities from
 *
 * http://kurucz.harvard.edu/grids/gridP00/ip00k2.pck19
 * (file dated 10-Mar-98 18:50)
 */ 
#include "FluxKurucz_35.h"
#include "FluxKurucz_45.h"
#include "FluxKurucz_40.h"
#include "FluxKurucz_50.h"

/* PHOENIX Intensities from Peter Hauschildt
 * Models below 3000K are 'dusty'
 */ 
#include "FluxHauschildt_35.h"
#include "FluxHauschildt_40.h"
#include "FluxHauschildt_45.h"
#include "FluxHauschildt_50.h"

/* Scaling factors between models and/or BB fluxes
 */
#define  ADJUST_BB_UNUSED 1
#include "FluxScales_50.h"
#include "FluxScales_45.h"
#include "FluxScales_40.h"
#include "FluxScales_35.h"
#undef   ADJUST_BB_UNUSED

struct flux
{
  float tmin_accept;
  int   g_index;
  int   len;
  float tlower;
  float tupper;

  float * flux;
  float * temp;
  float * adjust_factor;
};

struct flux fluxes[] =
  {
    { SWITCH_TEMP, 0, k_35_max, k_35_low, k_35_high, &k_flux_35[0][0], k_temp_35, adjust_flux_35 },
    { 0.0,         0, h_35_max, h_35_low, h_35_high, &h_flux_35[0][0], h_temp,    NULL },
    { SWITCH_TEMP, 1, k_40_max, k_40_low, k_40_high, &k_flux_40[0][0], k_temp_40, adjust_flux_40 },
    { 0.0,         1, h_40_max, h_40_low, h_40_high, &h_flux_40[0][0], h_temp,    NULL },
    { SWITCH_TEMP, 3, k_50_max, k_50_low, k_50_high, &k_flux_50[0][0], k_temp_50, adjust_flux_50 },
    { 0.0,         3, h_50_max, h_50_low, h_50_high, &h_flux_50[0][0], h_temp,    NULL },
    { SWITCH_TEMP, 2, k_45_max, k_45_low, k_45_high, &k_flux_45[0][0], k_temp_45, adjust_flux_45 },
    { 0.0,         2, h_45_max, h_45_low, h_45_high, &h_flux_45[0][0], h_temp,    NULL }
  };

void readFluxes(int pos)
{
  int nmodel;
  
  for (nmodel = 0; nmodel < 8; ++nmodel)
    {
      struct flux * model = &(fluxes[nmodel]);
      struct model_flux * mFlux = calloc(1, sizeof(struct model_flux));
      unsigned int i;
      unsigned int ncols = model->len;
      
      mFlux->tmin    = (model->tmin_accept  < model->tlower) ? model->tlower : model->tmin_accept;
      mFlux->tmax    = (model->tmin_accept == (float)0.0   ) ? SWITCH_TEMP   : model->tupper;
      mFlux->g_index = model->g_index;
      mFlux->log_g   = LogG[model->g_index];
      mFlux->m_grid_n    = 0;
      mFlux->m_grid_nmax = 0;
      mFlux->m_grid_t    = NULL;
      mFlux->m_grid_f    = NULL;
      mFlux->next        = NULL;

      for (i = 0; i < ncols; ++i)
	{
	  unsigned int offset = pos*ncols;
	  
	  if (mFlux->m_grid_n == mFlux->m_grid_nmax)
	    extendFlux(mFlux);
	  
	  mFlux->m_grid_t[i] = model->temp[i];
	  mFlux->m_grid_f[i] = *(model->flux + offset + i);
	  if (model->adjust_factor)
	    mFlux->m_grid_f[i] *= model->adjust_factor[pos];
	  ++mFlux->m_grid_n;
	}
      
      mFlux->next         = pb_bands[pos].mlist;
      pb_bands[pos].mlist = mFlux;
    }
  return;
}

void readPassband(int pos)
{
  int i, j, k, l;
  int nmax;
  
  strncpy(pb_bands[pos].name, pb_name[pos], NF_PB_NMAX-1);
  pb_bands[pos].name[NF_PB_NMAX-1] = '\0';

  pb_bands[pos].wavelength = pb_wavel[pos];
  pb_bands[pos].m2nd       = pb_m2nd[pos];

  nmax = sizeof(TH)/sizeof(float);
  i = 0;
    
  for (j = 0; j < nmax; ++j)
    {
      if (TH[j] <= SWITCH_TEMP)
	{
	  if (pb_bands[pos].ld_grid_n[0][0] == pb_bands[pos].ld_grid_nmax[0][0])
	    for (k = 0; k < NF_PB_GMAX; ++k)
	      for (l = 0; l < NF_LDMAX; ++l)
		extendLDC(pos, k, l);

	  for (k = 0; k < NF_PB_GMAX; ++k)
	    for (l = 0; l < NF_LDMAX; ++l)
	      pb_bands[pos].ld_grid_t[k][l][i]   = TH[j];
	  
	  pb_bands[pos].ld_grid_ldc[0][0][i]      = hL_35[pos][j];
	  pb_bands[pos].ld_grid_ldc[1][0][i]      = hL_40[pos][j];
	  pb_bands[pos].ld_grid_ldc[2][0][i]      = hL_45[pos][j];
	  pb_bands[pos].ld_grid_ldc[3][0][i]      = hL_50[pos][j];
	  
	  pb_bands[pos].ld_grid_ldc[0][1][2*i]    = hQ1_35[pos][j];
	  pb_bands[pos].ld_grid_ldc[0][1][2*i+1]  = hQ2_35[pos][j];
	  pb_bands[pos].ld_grid_ldc[1][1][2*i]    = hQ1_40[pos][j];
	  pb_bands[pos].ld_grid_ldc[1][1][2*i+1]  = hQ2_40[pos][j];
	  pb_bands[pos].ld_grid_ldc[2][1][2*i]    = hQ1_45[pos][j];
	  pb_bands[pos].ld_grid_ldc[2][1][2*i+1]  = hQ2_45[pos][j];
	  pb_bands[pos].ld_grid_ldc[3][1][2*i]    = hQ1_50[pos][j];
	  pb_bands[pos].ld_grid_ldc[3][1][2*i+1]  = hQ2_50[pos][j];
	  
	  pb_bands[pos].ld_grid_ldc[0][2][2*i]   = hS1_35[pos][j];
	  pb_bands[pos].ld_grid_ldc[0][2][2*i+1] = hS2_35[pos][j];
	  pb_bands[pos].ld_grid_ldc[1][2][2*i]   = hS1_40[pos][j];
	  pb_bands[pos].ld_grid_ldc[1][2][2*i+1] = hS2_40[pos][j];
	  pb_bands[pos].ld_grid_ldc[2][2][2*i]   = hS1_45[pos][j];
	  pb_bands[pos].ld_grid_ldc[2][2][2*i+1] = hS2_45[pos][j];
	  pb_bands[pos].ld_grid_ldc[3][2][2*i]   = hS1_50[pos][j];
	  pb_bands[pos].ld_grid_ldc[3][2][2*i+1] = hS2_50[pos][j];

	  for (k = 0; k < NF_PB_GMAX; ++k)
	    for (l = 0; l < NF_LDMAX; ++l)
	      ++pb_bands[pos].ld_grid_n[k][l];
	  ++i;
	}
    }

  nmax = sizeof(TK)/sizeof(float);

  for (j = 0; j < nmax; ++j)
    {
      if (TK[j] > SWITCH_TEMP)
	{
	  if (pb_bands[pos].ld_grid_n[0][0] == pb_bands[pos].ld_grid_nmax[0][0])
	    for (k = 0; k < NF_PB_GMAX; ++k)
	      for (l = 0; l < NF_LDMAX; ++l)
		extendLDC(pos, k, l);

	  for (k = 0; k < NF_PB_GMAX; ++k)
	    for (l = 0; l < NF_LDMAX; ++l)
	      pb_bands[pos].ld_grid_t[k][l][i]   = TK[j];
	  
	  pb_bands[pos].ld_grid_ldc[0][0][i]      = kL_35[pos][j];
	  pb_bands[pos].ld_grid_ldc[1][0][i]      = kL_40[pos][j];
	  pb_bands[pos].ld_grid_ldc[2][0][i]      = kL_45[pos][j];
	  pb_bands[pos].ld_grid_ldc[3][0][i]      = kL_50[pos][j];
	  
	  pb_bands[pos].ld_grid_ldc[0][1][2*i]    = kQ1_35[pos][j];
	  pb_bands[pos].ld_grid_ldc[0][1][2*i+1]  = kQ2_35[pos][j];
	  pb_bands[pos].ld_grid_ldc[1][1][2*i]    = kQ1_40[pos][j];
	  pb_bands[pos].ld_grid_ldc[1][1][2*i+1]  = kQ2_40[pos][j];
	  pb_bands[pos].ld_grid_ldc[2][1][2*i]    = kQ1_45[pos][j];
	  pb_bands[pos].ld_grid_ldc[2][1][2*i+1]  = kQ2_45[pos][j];
	  pb_bands[pos].ld_grid_ldc[3][1][2*i]    = kQ1_50[pos][j];
	  pb_bands[pos].ld_grid_ldc[3][1][2*i+1]  = kQ2_50[pos][j];
	  
	  pb_bands[pos].ld_grid_ldc[0][2][2*i]   = kS1_35[pos][j];
	  pb_bands[pos].ld_grid_ldc[0][2][2*i+1] = kS2_35[pos][j];
	  pb_bands[pos].ld_grid_ldc[1][2][2*i]   = kS1_40[pos][j];
	  pb_bands[pos].ld_grid_ldc[1][2][2*i+1] = kS2_40[pos][j];
	  pb_bands[pos].ld_grid_ldc[2][2][2*i]   = kS1_45[pos][j];
	  pb_bands[pos].ld_grid_ldc[2][2][2*i+1] = kS2_45[pos][j];
	  pb_bands[pos].ld_grid_ldc[3][2][2*i]   = kS1_50[pos][j];
	  pb_bands[pos].ld_grid_ldc[3][2][2*i+1] = kS2_50[pos][j];

	  for (k = 0; k < NF_PB_GMAX; ++k)
	    for (l = 0; l < NF_LDMAX; ++l)
	      ++pb_bands[pos].ld_grid_n[k][l];
	  ++i;
	}
    }
  pb_bands[pos].ld_types = (LD_LINEAR | LD_SQR | LD_SQRT);

  readFluxes(pos);
  finishModel(pos);

  return;
}

#if !defined(NDEBUG)

static int testFlux(int pos)
{
  int nmodel;
  
  for (nmodel = 0; nmodel < 8; ++nmodel)
    {
      struct flux * model = &(fluxes[nmodel]);
      float t, lo, hi, step;
      unsigned int i;
      float * farr = malloc(sizeof(float) *  model->len);

      for (i = 0; i < model->len; ++i) {
	farr[i] = *(model->flux + pos*model->len + i);
	if (model->adjust_factor != NULL)
	  farr[i] *= model->adjust_factor[pos];
      }
      lo   = (model->tlower < model->tmin_accept) ? model->tmin_accept : model->tlower;
      hi   = (model->tmin_accept == (float)0.0)   ? SWITCH_TEMP        : model->tupper;
      step = (hi-lo)/100.;
      t    = lo;

      /*
      if (nmodel == 7 && pos == 11)
	{
	  for (i = 0; i < model->len; ++i) {
	    printf("flux [%d] %8.1f %e\n", nmodel, model->temp[i], farr[i]);
	  }
	}
      */
      
      while (t < hi) {
	float * arr_t, * arr_f;
	int   ignore;
	float res1 = nf_interpolate(t, model->temp, farr, model->len, &ignore);
	int   len  = nf_modelFlux(pos, LogG[model->g_index], t,
				  &arr_t, &arr_f);
	float res2 = nf_interpolate(t, arr_t, arr_f, len, &ignore);
	if (fabs(res1 - res2) > 10*(res1+res2)*FLT_EPSILON) {
	  perr("flux mismatch [%d] %f %e <-> %e", nmodel, t, res1, res2);
	  for (i = 0; i < model->len; ++i)
	    printf("flux [%d] %8.1f %e\n", nmodel, model->temp[i], farr[i]);
	}
	/* printf("flux [%d] %8.1f %e <-> %e\n", nmodel, t, res1, res2); */

	t += step;
      }
    }
  return 0;
}

static int testLDC(int pos)
{
  float ldc[2];
  float * res;
  float *u, *a, *b, *c, *d;
  int i, j;
  const int ndata = 84;

  for (i = 0; i < NF_PB_GMAX; ++i)
    {
      if (ndata != pb_bands[pos].ld_grid_n[i][0]) {
	perr("wrong number of ldc values [%d] %d <-> %d", i, pb_bands[pos].ld_grid_n[i][0], ndata);
	return -1; }

      if (i == 0) {
	u = &kL_35[pos][0];  a = &kQ1_35[pos][0]; b = &kQ2_35[pos][0]; 
	c = &kS1_35[pos][0]; d = &kS2_35[pos][0];
      } else if (i == 1) {
	u = &kL_40[pos][0];  a = &kQ1_40[pos][0]; b = &kQ2_40[pos][0]; 
	c = &kS1_40[pos][0]; d = &kS2_40[pos][0];
      } else if (i == 2) {
	u = &kL_45[pos][0];  a = &kQ1_45[pos][0]; b = &kQ2_45[pos][0]; 
	c = &kS1_45[pos][0]; d = &kS2_45[pos][0];
      } else {
	u = &kL_50[pos][0];  a = &kQ1_50[pos][0]; b = &kQ2_50[pos][0]; 
	c = &kS1_50[pos][0]; d = &kS2_50[pos][0];
      }
      
      for (j = 0; j < (sizeof(TK)/sizeof(float)); ++j)
	{
	  if (TK[j] > SWITCH_TEMP)
	    {
	      res = nf_pb_fill_ldArray(pos, LD_LINEAR, LogG[i], TK[j], ldc);
	      if (res[0] != u[j]) {
		perr("ldc mismatch K linear [%d] %d %f %f %f", i, j, TK[j], res[0], u[j]);
		return -1; }
	      res = nf_pb_fill_ldArray(pos, LD_SQR, LogG[i], TK[j], ldc);
	      if (res[0] != a[j] || res[1] != b[j]) {
		perr("ldc mismatch K sqr [%d] %d %f (%f %f) (%f %f)", i, j, TK[j],
		     res[0], res[1], a[j], b[j]);
		return -1; }
	      res = nf_pb_fill_ldArray(pos, LD_SQRT, LogG[i], TK[j], ldc);
	      if (res[0] != c[j] || res[1] != d[j]) {
		perr("ldc mismatch K sqrt [%d] %d %f (%f %f) (%f %f)", i, j, TK[j],
		     res[0], res[1], c[j], d[j]);
		return -1; }
	    }
	}

      if (i == 0) {
	u = &hL_35[pos][0];  a = &hQ1_35[pos][0]; b = &hQ2_35[pos][0]; 
	c = &hS1_35[pos][0]; d = &hS2_35[pos][0];
      } else if (i == 1) {
	u = &hL_40[pos][0];  a = &hQ1_40[pos][0]; b = &hQ2_40[pos][0]; 
	c = &hS1_40[pos][0]; d = &hS2_40[pos][0];
      } else if (i == 2) {
	u = &hL_45[pos][0];  a = &hQ1_45[pos][0]; b = &hQ2_45[pos][0]; 
	c = &hS1_45[pos][0]; d = &hS2_45[pos][0];
      } else {
	u = &hL_50[pos][0];  a = &hQ1_50[pos][0]; b = &hQ2_50[pos][0]; 
	c = &hS1_50[pos][0]; d = &hS2_50[pos][0];
      }
      
      for (j = 0; j < (sizeof(TH)/sizeof(float)); ++j)
	{
	  if (TH[j] <= SWITCH_TEMP)
	    {
	      res = nf_pb_fill_ldArray(pos, LD_LINEAR, LogG[i], TH[j], ldc);
	      if (res[0] != u[j]) {
		perr("ldc mismatch K linear [%d] %d %f %f %f", i, j, TH[j], res[0], u[j]);
		return -1; }
	      res = nf_pb_fill_ldArray(pos, LD_SQR, LogG[i], TH[j], ldc);
	      if (res[0] != a[j] || res[1] != b[j]) {
		perr("ldc mismatch K sqr [%d] %d %f (%f %f) (%f %f)", i, j, TH[j],
		     res[0], res[1], a[j], b[j]);
		return -1; }
	      res = nf_pb_fill_ldArray(pos, LD_SQRT, LogG[i], TH[j], ldc);
	      if (res[0] != c[j] || res[1] != d[j]) {
		perr("ldc mismatch K sqrt [%d] %d %f (%f %f) (%f %f)", i, j, TH[j],
		     res[0], res[1], c[j], d[j]);
		return -1; }
	    }
	}
    }
  return 0;
}

static int testBB(int pos)
{
  int j;
  for (j = 0; j < NF_PB_GMAX; ++j)
    {   
      if (pb_bands[pos].t_upper[j] == (float)0.0) { perr("t adjust hi %d %d null", pos, j); return -1; }
      if (pb_bands[pos].t_lower[j] == (float)0.0) { perr("t adjust lo %d %d null", pos, j); return -1; }
      if (pb_bands[pos].bb_conv_upper[j] == (float)0.0) { perr("bb adjust hi %d %d null", pos, j); return -1; }
      if (pb_bands[pos].bb_conv_lower[j] == (float)0.0) { perr("bb adjust lo %d %d null", pos, j); return -1; }
    }
  return 0;
}

static void verifyPassband(int pos)
{
  int nerr = 0;

  if (0 != strcmp(pb_bands[pos].name, pb_name[pos])) {
    perr("** name mismatch %s <-> %s", pb_bands[pos].name, pb_name[pos]);
    ++nerr; }

  if (pb_bands[pos].wavelength != pb_wavel[pos]) {
    perr("** wavelength mismatch %f <-> %f", pb_bands[pos].wavelength, pb_wavel[pos]);
    ++nerr; }

  if (pb_bands[pos].m2nd != pb_m2nd[pos]) {
    perr("** 2nd moment mismatch %f <-> %f", pb_bands[pos].m2nd, pb_m2nd[pos]);
    ++nerr; }

  if (0 != pb_bands[pos].external) {
    perr("** external flag set");
    ++nerr; }

  if (pb_bands[pos].ld_types != (LD_LINEAR | LD_SQR | LD_SQRT)) {
    perr("** ldc types wrong");
    ++nerr; }

  if (0 != testLDC(pos)) {
    perr("** LDC values incorrect");
    ++nerr; }

  if (0 != testFlux(pos)) {
    perr("** Flux values incorrect");
    ++nerr; }

  if (0 != testBB(pos)) {
    perr("** BB adjustment values incorrect");
    ++nerr; }

  return;
}
/* if !defined(NDEBUG) */
#endif

/* Fill unused slots with compiled-in data
 */
void    nf_pb_readInternal()
{
  unsigned int pos;

  for (pos = 0; pos < NUM_MAG; ++pos)
    {
      if (pb_bands[pos].mlist == NULL)
	{
	  if (pb_bands[pos].external != 0)
	    {
	      /* External data, e.g. for limb darkening, but no model fluxes */
	      char msg[256];
	      snprintf(msg, sizeof(msg),
		       _("no flux data in external passband %d (%s)"),
		       pos,
		       (pb_bands[pos].name) ? pb_bands[pos].name : "NULL");
	      WARNING(msg);
	    }
	  else
	    {
	      /* No (valid) external data, use compiled-in data. */
	      readPassband(pos);
#if !defined(NDEBUG)
	      fprintf(stderr, "Using internal passband [%02d]\n", pos);
	      verifyPassband(pos);
#endif
	    }
	}
    }
  return;
}
