#ifndef LIGHT_PASSBAND_H
#define LIGHT_PASSBAND_H

#define NF_PB_NMAX FILTERNAME_SIZE

/* Number of supported log_g values */
#define NF_PB_GMAX  4

/* Maximum number of LDC coefficients */
#define NF_PB_LMAX  4

/* Number of LD Models */
#define NF_LDMAX    4


/* == FILE FORMAT ==
 *
 * Comments start with '#', empty/whitespace lines are ignored
 *
 * NAME 16c
 *
 * # Wavelength in micrometer (1000 A = 0.1 mu)
 * WAVELENGTH float
 *
 * # (second moment of passbands)^2 / FWHM
 * # may be approximated following the prescription by                          
 * # Young A.T. 1992, Astronomy and Astrophysics 257, 366; Equation 12:
 * # m2 = FullWidth(0.05 of peak)/4.4
 * # => (m2)^2/FWHM   
 * CORR_FACTOR float
 *
 * LDC_START
 * ld_t ld_log_g ld_lin ld_sqr_1 ld_sqr_2 ld_sqrt_1 ld_sqrt_2
 * 
 *
 * FLUX_START log_g f_tmin f_tmax
 * t flux
 * 
 */

typedef enum {
#define LD_LN 0
  LD_LINEAR = (1 << LD_LN),
#define LD_QD 1
  LD_SQR    = (1 << LD_QD),
#define LD_SQ 2
  LD_SQRT   = (1 << LD_SQ),
#define LD_4P 3
  LD_FOUR   = (1 << LD_4P),
} ldType;


/* ----------  Methods required ---------- */

/* Fill unused slots with compiled-in data
 */
void    nf_pb_readInternal();
void    nf_pb_updateNames(char filter[][NF_PB_NMAX]);

char  * nf_pb_get_name(unsigned int pos);
float   nf_pb_get_wave(unsigned int pos);
void    nf_pb_set_wave(unsigned int pos, float wave);
float   nf_pb_get_m2nd(unsigned int pos);

unsigned int nf_pb_has_ldType(int pos, ldType type);

/* Caller must ensure that 'array' is big enough for ldType;
 * Returns: pointer to 'array', or NULL on failure 
 */
float * nf_pb_fill_ldArray(int pos, ldType type, float log_g, float t,
			   float * array);

/* Return bb adjustment value (flux = bb_flux * adjust)
 */
double   nf_bb_adjust(int pos, float log_g, float t);

/* Returns size of array(s), or 0 on failure
 */
unsigned int nf_modelFlux(int pos, float log_g, float t,
			  float ** tarr, float ** farr);

/* Returns interpolated value
 */
float nf_interpolate(float t, float * tarr, float * farr, int len, int * where);

/* Exits on failure. fileName is "pos:path"
 */
void     nf_pb_readFile(const char * fileName);

#endif
